use std::env;
use std::path::Path;
use std::ffi::OsStr;
use std::fs;
use std::fs::DirBuilder;

///Creates directory (and any parents) and handles any errors that may occur
///Returns the following codes:
///OK: Successfully created the directory and its parents
///ERR: Unable to create the directory
pub fn make_dir<P: Into<String>>(path: P) -> String {
    let mut dir_build = DirBuilder::new();
    dir_build.recursive(true);

    match dir_build.create(path.into()) {
        Ok(_) => "OK".to_string(),
        Err(_) => "ERR".to_string()
    }
}


///abstraction around fs::read_to_string() that includes error handling
///Returns the file content as a `String` or returns "ERR" if something went wrong
pub fn read_file<P: Into<String>>(path: P) -> String {
    match fs::read_to_string(path.into()) {
        Ok(file) => file,
        Err(_) => "ERR".to_string()
    }
}

///abstraction around fs::write() that includes error handling
///Returns the following codes:
///OK: Save operation successful
///ERR: There was a problem writing the file
pub fn save_file<C: Into<String>, P: Into<String>>(path: P, contents: C) -> String {
    match fs::write(path.into(), contents.into()) {
        Ok(_) => "OK".to_string(),
        Err(_) => "ERR".to_string()
    }
}

///saves a string to a shared value (temp file in /tmp or ~/.cache)
///value must be a string value
pub fn set_shared_val<V: Into<String>>(name: &str, val: V) {
    let value = val.into();

    //get the .cache folder path
    let cache = vec![env::var("HOME").unwrap(),"/.cache/".to_string()];
    let cache_dir = concat_str(cache);

    let code = save_file(concat_str(vec!["/tmp/",name]), &value);
    if code == "ERR" {
        save_file(concat_str(vec![cache_dir,name.to_string()]), &value);
    }
    
}

///gets a shared value (temp file in /tmp or .cache)
///value must be a string value
pub fn get_shared_val(name: &str) -> String{
    //get the .cache folder path
    let cache = vec![env::var("HOME").unwrap(),"/.cache/".to_string()];
    let cache_dir = concat_str(cache);

    let mut val = read_file(concat_str(vec!["/tmp/",name]));

    if val == "ERR" {
        val = read_file(concat_str(vec![cache_dir,name.to_string()]));
    }
    val
}

///returns a list of files/directories using path string
pub fn get_dir<D: Into<String>>(dir: D, show_hidden: bool) -> Vec<String> {
    let dir_str = String::from(dir.into());
    let mut dir_name = dir_str.clone();
    if dir_str.chars().last().unwrap()!='/'{
        dir_name = concat_str(vec![dir_str.as_str(),"/"]);
    }

    let dir = fs::read_dir(&dir_name).unwrap();
    let mut output: Vec<String> = Vec::new();

    for file in dir {
        let name = file.unwrap();
        let sep = "/".to_string();
        let filename = get_filename(name.path().display().to_string());
        let attrs = name.metadata().unwrap();
        if show_hidden || filename.chars().nth(0).unwrap() != '.' {
            if attrs.is_dir() {
                let path_t = concat_str(vec![name.path().display().to_string(),sep]);
                let path: Vec<&str> = path_t.split('/').collect();
                let path2 = vec![path[path.len()-2].to_string(),"/".to_string()];
                output.push(concat_str(path2));
            }
            else {
                output.push(get_filename(name.path().display().to_string()));
            }  
        }
    }
    alphabetize(output)

    
}

///Alphabetizes a vector of strings
pub fn alphabetize<L: Into<String>>(list: Vec<L>) -> Vec<String>
where  L: Clone{
    //extract the strings from the generic value
    let mut string_list: Vec<String> = Vec::new();
    for item in list {
        string_list.push(item.into());
    }

    string_list.sort_by(|a, b| a.to_lowercase().cmp(&b.to_lowercase()));
    string_list
}

///get the name of the file/directory from a file path
pub fn get_filename<D: Into<String>>(dir: D) -> String {
    let dir_t =dir.into();
    let dir_vec = dir_t.split("/");
    dir_vec.last().unwrap().to_string()
}

///gets the parent directory from a file path
pub fn get_parent<D: Into<String>>(dir: D) -> String {
    let dir_t = dir.into();
    let mut dir_vec: Vec<&str> = dir_t.split("/").collect();
    dir_vec.pop();
    dir_vec.pop();
    let mut new_dir= String::from("");

    for (_, folder) in dir_vec.iter().enumerate(){
        for c in folder.chars(){
            new_dir.push(c);
        }
        new_dir.push('/');
    }
    new_dir
}

///capitalize the first letter in a string
pub fn capitalize<W: Into<String>>(word: W) -> String {
    let s1 = word.into();
    let mut v: Vec<char> = s1.chars().collect();

    for i in 0..s1.len(){
        if !v[i].is_digit(10){
            v[i] = v[i].to_uppercase().nth(0).unwrap();
            break;
        }
    }
    let s2: String = v.into_iter().collect();
    s2.to_string()
}

///returns the location (starting from 0) of the first instance of a character in a string
///skips any leading digits
///useful for parsing arguments
///returns the string's length if the char is not found
pub fn find_char<T: Into<String>>(text: T, sym: char) -> usize {
    let mut counter = 0;
    
    for c in text.into().chars(){
        if c == sym{
            break;
        }
        counter += 1;
    }

    counter
}

///concatenates (joins) 2 or more strings in a vector together
pub fn concat_str<S: Into<String>>(text: Vec<S>) -> String {
    let mut out = String::from("");

    for work_string in text {
        for c in work_string.into().chars(){
            out.push(c)
        }
    }
    
    out
}

///get the name of the executable
pub fn get_execname() -> String {
    env::args().next()
        .as_ref()
        .map(Path::new)
        .and_then(Path::file_name)
        .and_then(OsStr::to_str)
        .map(String::from).unwrap()
}